# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import pytest

from galaxy_importer import __version__
from galaxy_importer import main


def test_version():
    assert __version__ == '0.1.0'


def test_parser():
    parser = main.parse_args(['path/to/my_file.tar.gz'])
    assert parser.file == 'path/to/my_file.tar.gz'
    assert not parser.print_result

    parser = main.parse_args(['my_file.tar.gz', '--print-result'])
    assert parser.file == 'my_file.tar.gz'
    assert parser.print_result

    # SystemExit with missing required positional file argument
    with pytest.raises(SystemExit):
        main.parse_args(['--print-result'])



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import os
import pytest
import tempfile

from galaxy_importer.collection import CollectionLoader
from galaxy_importer.exceptions import ManifestValidationError


MANIFEST_JSON = """
{
 "collection_info": {
  "namespace": "my_namespace",
  "name": "my_collection",
  "version": "2.0.2",
  "authors": [
   "John Doe"
  ],
  "readme": "README.md",
  "tags": [
   "deployment",
   "fedora"
  ],
  "description": "A collection with various roles and plugins",
  "license": [
   "GPL-3.0-or-later",
   "MIT"
  ],
  "license_file": null,
  "dependencies": {
   "my_namespace.collection_nginx": ">=0.1.6",
   "network_user.collection_inspect": "2.0.0",
   "dave.deploy": "*"
  },
  "repository": null,
  "documentation": null,
  "homepage": null,
  "issues": null
 }
}
"""


def test_manifest_success():
    with tempfile.TemporaryDirectory() as temp_dir:
        with open(os.path.join(temp_dir, 'MANIFEST.json'), 'w') as fh:
            fh.write(MANIFEST_JSON)

        data = CollectionLoader(
            temp_dir, 'my_namespace-my_collection-2.0.2.tar.gz').load()
        assert data['metadata']['namespace'] == 'my_namespace'
        assert data['metadata']['name'] == 'my_collection'
        assert data['metadata']['version'] == '2.0.2'
        assert data['metadata']['authors'] == ['John Doe']
        assert data['metadata']['readme'] == 'README.md'
        assert data['metadata']['tags'] == ['deployment', 'fedora']
        assert data['metadata']['description'] == \
            'A collection with various roles and plugins'
        assert data['metadata']['license_file'] is None
        assert data['metadata']['dependencies'] == {
            'my_namespace.collection_nginx': '>=0.1.6',
            'network_user.collection_inspect': '2.0.0',
            'dave.deploy': '*'
        }
        assert data['metadata']['repository'] is None
        assert data['metadata']['homepage'] is None
        assert data['metadata']['issues'] is None
        assert data['error'] is None
        assert data['result'] == 'completed'


@pytest.mark.parametrize(
    'manifest_text,new_text,error_subset',
    [
        ('my_namespace', '', "'namespace' is required"),
        ('my_namespace', '00my.name.space', "'namespace' has invalid format"),
        ('my_collection', '', "'name' is required"),
        ('my_collection', '_my_collection', "'name' has invalid format"),
        ('2.0.2', '', "'version' is required"),
        ('2.0.2', '2.2.0.0.2', "semantic version format"),
        ('"John Doe"', '', "'authors' is required"),
        ('[\n   "John Doe"\n  ]', '"John Doe"', "to be a list of strings"),
        ('README.md', '', "'readme' is required"),
        ('"fedora"', '["fedora"]', "to be a list of strings"),
        ('"deployment",', '"tag",'*30, "Expecting no more than 20 tags"),
        ('"A collection with various roles and plugins"', '[]', "be a string"),
        ('"MIT"', '{}', "to be a list of strings"),
        ('"MIT"', '"not-a-valid-license-id"', "list of valid SPDX license"),
        ('"*"', '555', "Expecting depencency version to be string"),
        ('"dave.deploy"', '"davedeploy"', "Invalid dependency format:"),
        ('"dave.deploy"', '"007.deploy"', "Invalid dependency format: '007'"),
        ('"dave.deploy"', '"my_namespace.my_collection"', "self dependency"),
        ('"*"', '"3.4.0.4"', "version spec range invalid"),
        ('"repository": null', '"repository": []', "must be a string"),
        ('"documentation": null', '"documentation": []', "must be a string"),
        ('"homepage": null', '"homepage": []', "must be a string"),
        ('"issues": null', '"issues": []', "must be a string"),
    ],
)
def test_manifest_fail(manifest_text, new_text, error_subset):
    manifest_edited = MANIFEST_JSON.replace(manifest_text, new_text)
    with tempfile.TemporaryDirectory() as temp_dir:
        with open(os.path.join(temp_dir, 'MANIFEST.json'), 'w') as fh:
            fh.write(manifest_edited)

        with pytest.raises(ManifestValidationError,
                           match=error_subset):
            CollectionLoader(
                temp_dir, 'my_namespace-my_collection-2.0.2.tar.gz').load()



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import json
import re

import attr
import semantic_version

from galaxy_importer import constants
from galaxy_importer.utils.spdx_licenses import is_valid_license_id

SHA1_LEN = 40


def convert_none_to_empty_dict(val):
    """Returns an empty dict if val is None."""

    # if val is not a dict or val 'None' return val
    # and let the validators raise errors later
    if val is None:
        return {}
    return val


_FILENAME_RE = re.compile(
    r'^(?P<namespace>\w+)-(?P<name>\w+)-'
    r'(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$'
)


@attr.s(slots=True)
class CollectionFilename(object):

    namespace = attr.ib()
    name = attr.ib()
    version = attr.ib(converter=semantic_version.Version)

    def __str__(self):
        return f'{self.namespace}-{self.name}-{self.version}.tar.gz'

    @classmethod
    def parse(cls, filename):
        match = _FILENAME_RE.match(filename)
        if not match:
            raise ValueError(
                'Invalid filename. Expected: '
                '{namespace}-{name}-{version}.tar.gz'
            )

        return cls(**match.groupdict())

    @namespace.validator
    @name.validator
    def _validator(self, attribute, value):
        if not constants.NAME_REGEXP.match(value):
            raise ValueError(
                'Invalid {0}: {1!r}'.format(attribute.name, value)
            )


@attr.s(frozen=True)
class CollectionInfo(object):
    """Represents collection_info metadata in collection manifest."""

    namespace = attr.ib(default=None)
    name = attr.ib(default=None)
    version = attr.ib(default=None)
    license = attr.ib(factory=list)
    description = attr.ib(default=None)

    repository = attr.ib(default=None)
    documentation = attr.ib(default=None)
    homepage = attr.ib(default=None)
    issues = attr.ib(default=None)

    authors = attr.ib(factory=list)
    tags = attr.ib(factory=list)

    license_file = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(str)))
    readme = attr.ib(default=None)

    dependencies = attr.ib(
        factory=dict,
        converter=convert_none_to_empty_dict,
        validator=attr.validators.instance_of(dict))

    @property
    def label(self):
        return f"{self.namespace}.{self.name}"

    @staticmethod
    def value_error(msg):
        raise ValueError(f"Invalid collection metadata. {msg}") from None

    @namespace.validator
    @name.validator
    @version.validator
    @readme.validator
    @authors.validator
    def _check_required(self, attribute, value):
        """Check that value is present."""
        if not value:
            self.value_error(f"'{attribute.name}' is required")

    @namespace.validator
    @name.validator
    def _check_name(self, attribute, value):
        """Check value against name regular expression."""
        if not re.match(constants.NAME_REGEXP, value):
            self.value_error(f"'{attribute.name}' has invalid format: {value}")

    @version.validator
    def _check_version_format(self, attribute, value):
        """Check that version is in semantic version format."""
        if not semantic_version.validate(value):
            self.value_error(
                "Expecting 'version' to be in semantic version "
                f"format, instead found '{value}'.")

    @authors.validator
    @tags.validator
    @license.validator
    def _check_list_of_str(self, attribute, value):
        """Check that value is a list of strings."""
        err_msg = "Expecting '{attr}' to be a list of strings"
        if not isinstance(value, list):
            self.value_error(err_msg.format(attr=attribute.name))
        for list_item in value:
            if not isinstance(list_item, str):
                self.value_error(err_msg.format(attr=attribute.name))

    @license.validator
    def _check_licenses(self, attribute, value):
        """Check that all licenses in license list are valid."""
        invalid_licenses = [id for id in value if not is_valid_license_id(id)]
        if invalid_licenses:
            self.value_error(
                "Expecting 'license' to be a list of valid SPDX license "
                "identifiers, instead found invalid license identifiers: '{}' "
                "in 'license' value {}. "
                "For more info, visit https://spdx.org"
                .format(', '.join(invalid_licenses), value))

    @dependencies.validator
    def _check_dependencies_format(self, attribute, dependencies):
        """Check type and format of dependencies collection and version."""
        for collection, version_spec in dependencies.items():
            if not isinstance(collection, str):
                self.value_error("Expecting depencency to be string")
            if not isinstance(version_spec, str):
                self.value_error("Expecting depencency version to be string")

            try:
                namespace, name = collection.split('.')
            except ValueError:
                self.value_error(f"Invalid dependency format: '{collection}'")

            for value in [namespace, name]:
                if not re.match(constants.NAME_REGEXP, value):
                    self.value_error(
                        f"Invalid dependency format: '{value}' "
                        f"in '{namespace}.{name}'")

            if namespace == self.namespace and name == self.name:
                self.value_error("Cannot have self dependency")

            try:
                semantic_version.Spec(version_spec)
            except ValueError:
                self.value_error(
                    "Dependency version spec range invalid: "
                    f"{collection} {version_spec}")

    @tags.validator
    def _check_tags(self, attribute, value):
        """Check max tags and check against tag regular expression."""
        if value is not None and len(value) > constants.MAX_TAGS_COUNT:
            self.value_error(
                f"Expecting no more than {constants.MAX_TAGS_COUNT} tags "
                "in metadata")
        for tag in value:
            if not re.match(constants.TAG_REGEXP, tag):
                self.value_error(f"'tag' has invalid format: {tag}")

    @description.validator
    @repository.validator
    @documentation.validator
    @homepage.validator
    @issues.validator
    def _check_non_null_str(self, attribute, value):
        """Check that if value is present, it must be a string."""
        if value is not None and not isinstance(value, str):
            self.value_error(f"'{attribute.name}' must be a string")

    def __attrs_post_init__(self):
        """Checks called post init validation."""
        self._check_license_or_license_file(self.license, self.license_file)

    def _check_license_or_license_file(self, license_ids, license_file):
        """Confirm presence of either license or license_file."""
        if license_ids or license_file:
            return
        self.value_error(
            "Valid values for 'license' or 'license_file' are required. "
            f"But 'license' ({license_ids}) and "
            f"'license_file' ({license_file}) were invalid.")


@attr.s(frozen=True)
class CollectionArtifactManifest(object):
    """Represents collection manifest metadata."""

    collection_info = attr.ib(type=CollectionInfo)
    format = attr.ib(default=1)
    file_manifest_file = attr.ib(factory=dict)

    @classmethod
    def parse(cls, data):
        meta = json.loads(data)
        col_info = meta.pop('collection_info', None)
        meta['collection_info'] = CollectionInfo(**col_info)
        return cls(**meta)


@attr.s(frozen=True)
class ImportResult(object):
    """Result of the import process, collection metadata, and contents."""

    metadata = attr.ib(default=None, type=CollectionInfo)
    docs_blob = attr.ib(factory=dict)
    contents = attr.ib(factory=list)
    custom_license = attr.ib(default=None)
    result = attr.ib(default=None)
    error = attr.ib(default=None)



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import argparse
import json
import logging
import sys

from galaxy_importer import collection


def main(args=None):
    logging.basicConfig(
        stream=sys.stdout,
        format='%(levelname)s: %(message)s',
        level=logging.INFO)
    args = parse_args(args)

    data = call_importer(file=args.file)

    if args.print_result:
        print(json.dumps(data, indent=4))

    write_output_file(data)


def parse_args(args):
    parser = argparse.ArgumentParser(
        description='Run importer on collection and save result to disk.')
    parser.add_argument(
        'file',
        help='artifact to import')
    parser.add_argument(
        '--print-result',
        dest='print_result',
        action='store_true',
        help='print importer result to console')
    return parser.parse_args(args=args)


def call_importer(file):
    """Returns result of galaxy_importer import process.

    :param file: Artifact file to import.
    """
    data = collection.import_collection(file)
    if data['result'] == 'completed':
        print('Importer processing completed successfully')
    else:
        print(f'Error during importer proccessing: {data["error"]}')
    return data


def write_output_file(data):
    with open('importer_result.json', 'w') as output_file:
        output_file.write(json.dumps(data))


if __name__ == '__main__':
    exit(main())



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import logging
import os
import tarfile
import tempfile

import attr

from galaxy_importer import exceptions as exc
from galaxy_importer import schema


default_logger = logging.getLogger(__name__)

ALLOWED_TYPES = ['text/markdown', 'text/x-rst']
RESULT_COMPLETED = 'completed'
RESULT_FAILED = 'failed'


def import_collection(filepath, logger=None):
    logger = logger or default_logger
    try:
        return _import_collection(filepath, logger)
    except Exception as exc:
        import_result = schema.ImportResult(
            result=RESULT_FAILED,
            error=str(exc),
        )
        return attr.asdict(import_result)


def _import_collection(filepath, logger):
    filename = os.path.basename(filepath)

    with tempfile.TemporaryDirectory() as extract_dir:
        with tarfile.open(filepath, 'r') as pkg_tar:
            pkg_tar.extractall(extract_dir)

        return CollectionLoader(extract_dir, filename, logger=logger).load()


class CollectionLoader(object):
    """Loads collection and content info."""

    def __init__(self, path, filename, logger=None):
        self.log = logger or default_logger
        self.path = path
        self.filename = filename

        self.metadata = None
        self.docs_blob = None
        self.contents = None

    def load(self):
        self._load_collection_manifest()
        # TODO(awcrosby): add filename check when worker can pass filename with
        # collection details instead of filename made of hash string

        import_result = schema.ImportResult(
            metadata=self.metadata,
            docs_blob=self.docs_blob,
            contents=self.contents,
            result=RESULT_COMPLETED,
            error=None,
        )
        return attr.asdict(import_result)

    def _load_collection_manifest(self):
        manifest_file = os.path.join(self.path, 'MANIFEST.json')
        if not os.path.exists(manifest_file):
            raise exc.ManifestNotFound('No manifest found in collection')

        with open(manifest_file, 'r') as f:
            try:
                data = schema.CollectionArtifactManifest.parse(f.read())
            except ValueError as e:
                raise exc.ManifestValidationError(str(e))
            self.metadata = data.collection_info



# (c) 2012-2019, Ansible by Red Hat
#
# This file is part of Ansible Galaxy
#
# Ansible Galaxy is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by
# the Apache Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Ansible Galaxy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# Apache License for more details.
#
# You should have received a copy of the Apache License
# along with Galaxy.  If not, see <http://www.apache.org/licenses/>.

import pytest
import semantic_version

from galaxy_importer import schema

NAMESPACE = 'my_namespace'
NAME = 'my_collection'
VERSION_STR = '4.3.0'

FILENAME_DATA = {
    'namespace': NAMESPACE,
    'name': NAME,
    'version': VERSION_STR,
}


def test_filename_parse():
    filename_str = f'{NAMESPACE}-{NAME}-{VERSION_STR}.tar.gz'
    filename_from_str = schema.CollectionFilename.parse(filename_str)

    filename_from_dict = schema.CollectionFilename(**FILENAME_DATA)

    for filename_obj in [filename_from_str, filename_from_dict]:
        assert str(filename_obj) == filename_str
        assert filename_obj.namespace == NAMESPACE
        assert filename_obj.name == NAME
        assert filename_obj.version == semantic_version.Version(VERSION_STR)


@pytest.mark.parametrize(
    'filename_str',
    [
        f'{NAMESPACE}-{NAME}-{VERSION_STR}.tar.gz',
        f'{NAMESPACE}-{NAME}-0.1.1-alpha+build.2012-05-15.tar.gz',
        pytest.param(f'{NAMESPACE}{NAME}{VERSION_STR}.tar.gz',
                     marks=pytest.mark.xfail),  # 0 dashes
        pytest.param(f'{NAMESPACE}-{NAME}{VERSION_STR}.tar.gz',
                     marks=pytest.mark.xfail),  # 1 dash
        pytest.param(f'{NAMESPACE}-{NAME}-{VERSION_STR}!.tar.gz',
                     marks=pytest.mark.xfail),  # bang
    ],
)
def test_filename_format(filename_str):
    res = schema.CollectionFilename.parse(filename_str)
    assert type(res) == schema.CollectionFilename


@pytest.mark.parametrize(
    'filename_str,error_subset',
    [
        (f'no__dunder-{NAME}-{VERSION_STR}.tar.gz', 'Invalid namespace:'),
        (f'{NAMESPACE}-0startswithnum-{VERSION_STR}.tar.gz', 'Invalid name:'),
    ]
)
def test_bad_name_error(filename_str, error_subset):
    with pytest.raises(ValueError) as excinfo:
        schema.CollectionFilename.parse(filename_str)
    assert error_subset in str(excinfo.value)
